import logging
import itertools
import operator
from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from .models import Action, Diff
from django.utils.html import format_html_join
from django.utils.safestring import mark_safe
from django.utils.html import conditional_escape
from django.contrib.admin.utils import unquote
from django.contrib.contenttypes.models import ContentType


logger = logging.getLogger(__name__)


def get_content_type_for_model(obj):
    return ContentType.objects.get_for_model(obj, for_concrete_model=False)


class HistoryBlockAdmin(object):

    def history_view(self, request, object_id, extra_context=None):
        model = self.model
        opts = model._meta
        action_list = Action.objects.filter(
            consumer_pk=unquote(object_id),
            consumer_type=get_content_type_for_model(model),
        ).order_by('-created_at')
        if self.inlines:
            for inline in self.inlines:
                for f in inline.model._meta.local_fields:
                    if f.remote_field and f.remote_field.model == model:
                        logger.info(f"{f.name}")
                        qs = {f"{f.name}_id": object_id}
                        ids = [obj.pk for obj in inline.model.objects.filter(**qs)]
                        action_list = itertools.chain(
                            action_list,
                            Action.objects.filter(
                                consumer_pk__in=ids,
                                consumer_type=get_content_type_for_model(inline.model),
                            ).order_by('-created_at')
                        )
            action_list = list(action_list)
        extra_context = {
            'action_list': action_list,
            'opts': opts,
            'history': Action._meta,
            **(extra_context or {}),
        }
        self.object_history_template = "history/admin/list.html"

        request.current_app = self.admin_site.name
        return super().history_view(request, object_id, extra_context=extra_context)


class DiffAdminInlines(admin.StackedInline):
    model = Diff
    exclude = ['change']
    extra = 0
    max_num = 0
    can_delete = False
    readonly_fields = ('version', 'change_html', 'field', 'field_name')

    def change_html(self, instance):
        return format_html_join(
            mark_safe('<br>'),
            '<span style="{}">{}</span>',
            (
                ('color:#cc0000;' if line.startswith('-') else
                 ('color:#008800;' if line.startswith('+') else
                  ('color:#990099;' if line.startswith('@') else '')),
                 conditional_escape(line)) for line in instance.change.splitlines()
            )
        )

    change_html.short_description = _("Изменения")

    def field_name(self, instance):
        return instance.verbose_field_name()

    field_name.short_description = _('Название поля')


@admin.register(Action)
class ActionAdmin(admin.ModelAdmin):
    list_display = ('__str__', 'profile', 'ip', 'diffs', 'created_at')
    readonly_fields = (
        'consumer_type', 'consumer_pk', 'consumer', 'action_type', 'profile', 'rollback_to', 'ip', 'show_in_timeline',
        'created_at'
    )
    inlines = [DiffAdminInlines]

    def diffs(self, obj):
        if obj.diff_set.all():
            return format_html_join(
                mark_safe('<br>'),
                '{}',
                ((diff,) for diff in obj.diff_set.all())
            )
        else:
            return "-"

    diffs.short_description = _("Изменения")

    def has_delete_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request):
        return False

    def change_view(self, request, object_id, extra_context=None):
        extra_context = extra_context or {}
        extra_context['show_save_and_continue'] = False
        extra_context['show_save'] = False
        return super(ActionAdmin, self).change_view(request, object_id, extra_context=extra_context)



import logging
import datetime
import difflib
from django.contrib.contenttypes.models import ContentType
from django.contrib.admin.models import LogEntry
from django.db import transaction
from . import defaults, utils
from .models import Diff, Action
from .middleware import get_current_ip, get_current_user

logger = logging.getLogger(__name__)


def get_last_version(consumer_type, consumer_pk, field):
    try:
        return Diff.objects.filter(
            action__consumer_type=consumer_type,
            action__consumer_pk=consumer_pk, field=field
        ).order_by('-version')[0].version
    except IndexError:
        return 0


@transaction.atomic
def _create_diff(sender_name, consumer_type, **kwargs):
    instance = kwargs.get('instance')

    if hasattr(instance, '_action_type'):
        action_type = instance._action_type
        delattr(instance, '_action_type')
    else:
        action_type = (
            defaults.ACTION_CREATE if kwargs.get('created') else
            defaults.ACTION_EDIT)

    if action_type == defaults.ACTION_ROLLBACK:
        return

    timeline_observed = utils.is_observed(sender_name, action_type)
    if not ((sender_name in defaults.OBSERVED_FIELDS) or timeline_observed):
        return

    profile = get_current_user()

    if action_type == defaults.ACTION_EDIT:
        if Action.objects.filter(
            consumer_type=consumer_type, consumer_pk=instance.pk,
            action_type__in=(defaults.ACTION_CREATE, defaults.ACTION_EDIT),
            created_at__gte=datetime.datetime.now()
        ).exists():
            instance._ignore_in_timeline = True
    try:
        _ignore_in_timeline = getattr(instance, '_ignore_in_timeline', False)
    except ValueError as e:
        logger.info("Ignore in timeline: {}".format(e))
        _ignore_in_timeline = False

    try:
        provider = Action.objects.create(
            consumer_type=consumer_type, consumer_pk=instance.pk,
            action_type=action_type, profile=profile, ip=get_current_ip(),
            show_in_timeline=_ignore_in_timeline)
    except ValueError as e:
        logger.warning("Value Error: {}".format(e))
        return

    if action_type == defaults.ACTION_CREATE:
        instance._ignore_in_timeline = True

    fields = (
        set(defaults.OBSERVED_FIELDS.get(sender_name, ()))
    )
    if fields:
        consumer_name = f"{provider.consumer}"

    has_diff = False

    for field in fields:
        last_version = get_last_version(
            consumer_type, instance.pk, field)

        if last_version > 0:
            diffs = Diff.objects.filter(
                action__consumer_type=consumer_type,
                action__consumer_pk=instance.pk, field=field,
                version=last_version)
            if len(diffs) > 1:
                for diff in diffs[1:]:
                    diff.delete()
            last_diff = diffs[0]
            last_value = last_diff.get_version_text()

        else:
            last_value = ''

        new_value = utils.to_string(getattr(instance, field))

        if last_value != new_value:
            try:
                last_action = Action.objects.filter(
                    diff__version=last_version, diff__field=field,
                    consumer_type=consumer_type,
                    consumer_pk=instance.pk).first()
                if last_action:
                    last_date = last_action.created_at
                else:
                    last_date = ''
            except IndexError as e:
                logger.warning("Last action: {}".format(e))
                last_date = ''
                last_value = ''

            patch = '\n'.join(difflib.unified_diff(
                last_value.splitlines(), new_value.splitlines(),
                consumer_name, consumer_name,
                str(last_date), str(provider.created_at),
                lineterm=''))

            if patch:
                has_diff = True
            else:
                continue

            new_version = last_version + 1
            diff = Diff.objects.create(
                action=provider, version=new_version, field=field,
                change=patch)

            # Make sure that we can get latest version without errors.
            diff = Diff.objects.get(
                action__consumer_type=consumer_type,
                action__consumer_pk=instance.pk, field=field,
                version=new_version)
            diff.get_version_text()

    if not has_diff and action_type == defaults.ACTION_EDIT:
        provider.delete()


def object_m2m_save(sender, **kwargs):
    sender_name = f"{sender._meta.app_label}.{sender._meta.object_name.split('_')[0]}"
    try:
        consumer_type = ContentType.objects.get(
            app_label=sender._meta.app_label,
            model=sender._meta.object_name.split('_')[0].lower()
        )
    except Exception as e:
        logger.error(f"Content Type not found: {e}")
        return

    _create_diff(sender_name, consumer_type, **kwargs)


def object_post_save(sender, **kwargs):
    sender_name = f"{sender._meta.app_label}.{sender._meta.object_name}"

    try:
        consumer_type = ContentType.objects.get_for_model(sender)
    except Exception as e:
        logger.error(f"Content Type not found: {e}")
        return

    _create_diff(sender_name, consumer_type, **kwargs)


def object_pre_delete(sender, **kwargs):
    sender_name = '.'.join((sender._meta.app_label, sender._meta.object_name))
    if (
        (sender_name in defaults.OBSERVED_FIELDS) or
        utils.is_observed(sender_name, defaults.ACTION_DELETE)):
        instance = kwargs.get('instance')
        Action.objects.create(
            consumer_type=ContentType.objects.get_for_model(sender),
            consumer_pk=instance.pk, action_type=defaults.ACTION_DELETE,
            profile=get_current_user(), ip=get_current_ip())



# Generated by Django 3.1.6 on 2021-06-22 05:34

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    operations = [
        migrations.CreateModel(
            name='Action',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('consumer_pk', models.PositiveIntegerField(blank=True, null=True, verbose_name='Consumer PK')),
                ('action_type', models.PositiveSmallIntegerField(choices=[(0, 'Созданный объект'), (1, 'Отредактированный объект'), (2, 'Удаленный объект'), (3, 'отмена'), (4, 'Переименованный объект'), (5, 'Перемещено в корзину'), (6, 'Восстановлен из корзины')], verbose_name='Тип действия')),
                ('ip', models.CharField(max_length=15, null=True, verbose_name='IP адрес')),
                ('show_in_timeline', models.BooleanField(default=True, verbose_name='Показать на временной шкале')),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now, verbose_name='создан')),
                ('consumer_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='content_type_set_for_action', to='contenttypes.contenttype', verbose_name='Consumer Type')),
                ('profile', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL, verbose_name='Профиль')),
            ],
            options={
                'verbose_name': 'действие',
                'verbose_name_plural': 'Действия',
                'ordering': ['-created_at'],
            },
        ),
        migrations.CreateModel(
            name='Diff',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('version', models.PositiveIntegerField(verbose_name='Версия')),
                ('change', models.TextField()),
                ('field', models.CharField(blank=True, max_length=255, null=True, verbose_name='Поле')),
                ('action', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='django_history.action')),
            ],
            options={
                'verbose_name': 'отличие',
                'verbose_name_plural': 'Отличия',
                'unique_together': {('action', 'field')},
                'index_together': {('version', 'field')},
            },
        ),
        migrations.AddField(
            model_name='action',
            name='rollback_to',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='rolled_back_from', to='django_history.diff', verbose_name='Откатить до'),
        ),
        migrations.AlterIndexTogether(
            name='action',
            index_together={('consumer_type', 'consumer_pk')},
        ),
    ]



from django.utils import timezone
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import gettext_lazy as _

from . import defaults, utils

User = get_user_model()


class Action(models.Model):
    consumer_type = models.ForeignKey(
        ContentType,
        verbose_name=_("Consumer Type"),
        related_name="content_type_set_for_%(class)s",
        blank=True, null=True, on_delete=models.SET_NULL,
    )
    consumer_pk = models.PositiveIntegerField(_("Consumer PK"), blank=True, null=True)
    consumer = GenericForeignKey(
        ct_field="consumer_type", fk_field="consumer_pk", )
    action_type = models.PositiveSmallIntegerField(
        _('Тип действия'), choices=defaults.ACTION_TYPES)
    profile = models.ForeignKey(
        User, null=True, verbose_name=_('Профиль'), on_delete=models.SET_NULL
    )
    rollback_to = models.ForeignKey(
        'Diff', null=True, related_name='rolled_back_from',
        verbose_name=_('Откатить до'), on_delete=models.SET_NULL
    )
    ip = models.CharField(_('IP адрес'), max_length=255, null=True)
    show_in_timeline = models.BooleanField(
        _('Показать на временной шкале'), default=True)
    created_at = models.DateTimeField(_('создан'), default=timezone.now)

    def is_rollback(self):
        return self.action_type == defaults.ACTION_ROLLBACK

    def consumer_model_verbose(self):
        return self.consumer_type.model_class()._meta.verbose_name

    def __str__(self):
        return f'{self.get_action_type_display()} {self.consumer}'

    class Meta:
        verbose_name = _("действие")
        verbose_name_plural = _("Действия")
        ordering = ['-created_at']
        index_together = ('consumer_type', 'consumer_pk')


class Diff(models.Model):
    action = models.ForeignKey(Action, on_delete=models.CASCADE)
    version = models.PositiveIntegerField(_("Версия"))
    change = models.TextField()
    field = models.CharField(
        max_length=255,
        verbose_name=_("Поле"),
        null=True, blank=True
    )

    def first_version(self):
        try:
            return Diff.objects.select_related('action').order_by(
                'version').get(
                action__consumer_type__id=self.action.consumer_type_id,
                action__consumer_pk=self.action.consumer_pk,
                field=self.field, version=1)
        except Diff.DoesNotExist:
            pass

    def prev_version(self):
        try:
            return Diff.objects.select_related('action').get(
                action__consumer_type__id=self.action.consumer_type_id,
                action__consumer_pk=self.action.consumer_pk,
                field=self.field, version=self.version - 1)
        except Diff.DoesNotExist:
            pass

    def next_version(self):
        try:
            return Diff.objects.select_related('action').get(
                action__consumer_type__id=self.action.consumer_type_id,
                action__consumer_pk=self.action.consumer_pk,
                field=self.field, version=self.version + 1)
        except Diff.DoesNotExist:
            pass

    def last_version(self):
        try:
            return Diff.objects.select_related('action').filter(
                action__consumer_type__id=self.action.consumer_type_id,
                action__consumer_pk=self.action.consumer_pk,
                field=self.field).order_by('-version')[0]
        except IndexError:
            pass

    def get_version_text(self):
        return utils.merge([
            value[0] for value in self.__class__.objects.filter(
                action__consumer_type__id=self.action.consumer_type_id,
                action__consumer_pk=self.action.consumer_pk, field=self.field,
                version__lte=self.version
            ).order_by('version').values_list('change', 'version')])

    def verbose_field_name(self):
        if self.field:
            model = self.action.consumer_type.model_class()
            name = getattr(
                getattr(
                    getattr(model, self.field, None), 'fget', None),
                'verbose_name', None)
            return (name or model._meta.get_field(
                self.field)).verbose_name

    def __str__(self):
        return f'{self.verbose_field_name()}@{self.get_version_text()}'

    class Meta:
        verbose_name = _("отличие")
        verbose_name_plural = _("Отличия")
        unique_together = (
            ('action', 'field'),
        )
        index_together = ('version', 'field')



# Generated by Django 3.1.6 on 2021-06-24 14:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_history', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='action',
            name='ip',
            field=models.CharField(max_length=255, null=True, verbose_name='IP адрес'),
        ),
    ]



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _
from django.db.models.signals import post_save, pre_delete, m2m_changed


class HistoryConfig(AppConfig):
    name = 'django_history'
    verbose_name = _('История изменений')

    def ready(self):
        from .signals import object_post_save, object_pre_delete, object_m2m_save

        post_save.connect(object_post_save)
        m2m_changed.connect(object_m2m_save)
        pre_delete.connect(object_pre_delete)

        return super().ready()



import itertools
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from .utils import observe


(
    ACTION_CREATE, ACTION_EDIT, ACTION_DELETE, ACTION_ROLLBACK,
    ACTION_RENAME, ACTION_TRASH, ACTION_RESTORE
) = range(0, 7)

ACTION_TYPES = (
    (ACTION_CREATE, _('Созданный объект')),
    (ACTION_EDIT, _('Отредактированный объект')),
    (ACTION_DELETE, _('Удаленный объект')),
    (ACTION_ROLLBACK, _('отмена')),
    (ACTION_RENAME, _('Переименованный объект')),
    (ACTION_TRASH, _('Перемещено в корзину')),
    (ACTION_RESTORE, _('Восстановлен из корзины')),
)

OBSERVED_FIELDS = getattr(
    settings, 'OBSERVED_FIELDS',
    {
        'accounts.User':
            (
                'username', 'first_name', 'middle_name', 'last_name', 'gender', 'phone', 'email',
                'photo', 'birthday',
            ),
     })

OBSERVED_FIELD_NAMES = list(set(itertools.chain.from_iterable(OBSERVED_FIELDS.values())))

TIMELINE_PAGE_DAYS = getattr(settings, 'TIMELINE_PAGE_DAYS', 7)

AUTHORS_TOP_LIMIT = getattr(settings, 'AUTHORS_TOP_LIMIT', 10)

TIMELINE_FEED_DAYS = getattr(settings, 'TIMELINE_FEED_DAYS', 3)



import datetime
import re
from django.db.models import Model, Q

HUNK_RE = '@@ \-(\d+)(,(\d+))? \+(\d+)(,(\d+))? @@'


def merge(diffs):
    """
    Apply multiple diffs.
    Add one line:
    >>> merge([
    ...     '--- \\n'
    ...     '+++ \\n'
    ...     '@@ -1,0 +1,1 @@\\n'
    ...     '+qwf'])
    u'qwf'
    Remove one line:
    >>> merge([
    ...     '--- \\n'
    ...     '+++ \\n'
    ...     '@@ -1,0 +1,1 @@\\n'
    ...     '+qwf\\n',
    ...     '--- \\n'
    ...     '+++ \\n'
    ...     '@@ -1,1 +1,0 @@\\n'
    ...     '-qwf'])
    u''
    Replace one line:
    >>> merge([
    ...     '--- \\n'
    ...     '+++ \\n'
    ...     '@@ -1,0 +1,2 @@\\n'
    ...     '+qwf\\n'
    ...     '+zxc\\n',
    ...     '--- \\n'
    ...     '+++ \\n'
    ...     '@@ -1,2 +1,2 @@\\n'
    ...     ' qwf\\n'
    ...     '-zxc\\n'
    ...     '+ars'])
    u'qwf\\nars'
    Mixed lines, multiple hunks:
    1234 -> 123456 -> 1357 -> 1839507abcd -> 893507abc
    >>> merge([
    ...     '--- \\n'
    ...     '+++ \\n'
    ...     '@@ -1,0 +1,4 @@\\n'
    ...     '+11\\n'
    ...     '+22\\n'
    ...     '+33\\n'
    ...     '+44',
    ...     '--- \\n'
    ...     '+++ \\n'
    ...     '@@ -2,3 +2,5 @@\\n'
    ...     ' 22\\n'
    ...     ' 33\\n'
    ...     ' 44\\n'
    ...     '+55\\n'
    ...     '+66',
    ...     '--- \\n'
    ...     '+++ \\n'
    ...     '@@ -1,6 +1,4 @@\\n'
    ...     ' 11\\n'
    ...     '-22\\n'
    ...     ' 33\\n'
    ...     '-44\\n'
    ...     ' 55\\n'
    ...     '-66\\n'
    ...     '+77',
    ...     '--- \\n'
    ...     '+++ \\n'
    ...     '@@ -1,4 +1,11 @@\\n'
    ...     ' 11\\n'
    ...     '+88\\n'
    ...     ' 33\\n'
    ...     '+99\\n'
    ...     ' 55\\n'
    ...     '+00\\n'
    ...     ' 77\\n'
    ...     '+aa\\n'
    ...     '+bb\\n'
    ...     '+cc\\n'
    ...     '+dd',
    ...     '--- \\n'
    ...     '+++ \\n'
    ...     '@@ -1,4 +1,3 @@\\n'
    ...     '-11\\n'
    ...     ' 88\\n'
    ...     ' 33\\n'
    ...     ' 99\\n'
    ...     '@@ -8,4 +7,3 @@\\n'
    ...     ' aa\\n'
    ...     ' bb\\n'
    ...     ' cc\\n'
    ...     '-dd'])
    u'88\\n33\\n99\\n55\\n00\\n77\\naa\\nbb\\ncc'
    Test with real world data:
    >>> merge([
    ... '--- [2011/05/12 13:22] zz\\n'
    ... '+++ [2011/05/12 13:22] zz 2011-05-12 13:22:51.895911\\n'
    ... '@@ -1,0 +1,1 @@\\n'
    ... '+<p>qqZZZ</p>',
    ... '--- [2011/05/12 13:22] zz 2011-05-12 13:22:51.895911\\n'
    ... '+++ [2011/05/12 13:22] zz 2011-05-12 13:23:06.504037\\n'
    ... '@@ -1,1 +1,3 @@\\n'
    ... ' <p>qqZZZ</p>\\n'
    ... '+<p>12323arsar</p>\\n'
    ... '+<p>arsars</p>',
    ... '--- [2011/05/12 13:22] zz 2011-05-12 13:23:06.504037\\n'
    ... '+++ [2011/05/12 13:22] zz 2011-05-12 13:23:24.784658\\n'
    ... '@@ -1,3 +1,1 @@\\n'
    ... ' <p>qqZZZ</p>\\n'
    ... '-<p>12323arsar</p>\\n'
    ... '-<p>arsars</p>'])
    u'<p>qqZZZ</p>'
    """
    hunk_re = re.compile(HUNK_RE)
    text = []
    for diff in diffs:
        text_new = text[:]
        lines = (line for line in diff.splitlines())
        line = ''
        while line is not None:
            try:
                line = lines.__next__()
                if line.startswith('--- '):
                    lines.__next__()
                    line = lines.__next__()

                # Parse hunk info.
                match = hunk_re.match(line)
                if match is None:
                    continue

                (start_old, has_size_old, size_old, start_new, has_size_new,
                 size_new) = match.groups()

                if has_size_old is None:
                    size_old = 1

                if has_size_new is None:
                    size_new = 1

                start_old, size_old, start_new, size_new = map(
                    int, (start_old, size_old, start_new, size_new))

                start_old -= 1
                start_new -= 1

                # Parse hunk.
                end_new = start_new + size_new
                end_old = start_old + size_old
                while start_new < end_new or start_old < end_old:
                    line = lines.__next__()

                    if line.startswith('-'):
                        # Line removed.
                        assert text[start_old] == line[1:], \
                            "Deleted text didn't match: %s\n%s" % (
                                text, line)
                        del text_new[start_new]
                        start_old += 1
                    elif line.startswith('+'):
                        # Line added.
                        text_new.insert(start_new, line[1:])
                        start_new += 1

                    else:
                        # Line unchanged.
                        line = line[1:]
                        assert text[start_old] == text_new[start_new] == line, \
                            "Text shouldn't change here: %s\n%s" % (
                                diff, (text[start_old], line, text))
                        start_old += 1
                        start_new += 1
            except Exception as e:
                line = None
        text = text_new

    return u'\n'.join(line for line in text if line is not None)


def to_string(value):
    if callable(value):
        value = ', '.join([val.__str__() for val in value.all()])
        return str(value)
    if isinstance(value, datetime.datetime):
        return f"{value.strftime('%Y-%m-%d %H:%M')}"
    elif isinstance(value, datetime.date):
        return f"{value.strftime('%Y-%m-%d')}"
    elif isinstance(value, Model):
        return f"{value.pk} ({value})"
    elif isinstance(value, str):
        return f"{value}"
    else:
        return f"{value}"


def from_string(value, to_type):
    if value is 'None':
        return None

    elif to_type:
        return value
    elif issubclass(to_type, str):
        return value.encode('utf-8')
    elif issubclass(to_type, bool):
        return value == u'True'
    elif issubclass(to_type, int):
        return int(value)
    elif issubclass(to_type, datetime.datetime):
        return datetime.datetime.strptime(value, '%d/%m/%Y %H:%M')
    elif issubclass(to_type, Model):
        return to_type.objects.get(pk=value.split(' ', 1)[0])
    else:
        raise NotImplementedError


_TIMELINE_FIELDS_DATA = {}


def observe(
    model, create=False, edit=False, delete=False, rollback=False,
    rename=False):
    assert any((create, edit, delete, rollback, rename)), \
        "At least one action type must be set"
    _TIMELINE_FIELDS_DATA[model] = (create, edit, delete, rollback, rename)
    return model


def is_observed(model, event):
    model_data = _TIMELINE_FIELDS_DATA.get(model)
    if not model_data:
        return False

    return model_data[event]


def get_consumer_filter(key):
    from . import defaults
    action_types = [
        action_type for action_type, title in defaults.ACTION_TYPES
        if _TIMELINE_FIELDS_DATA[key][action_type]]
    app, model = key.split('.')
    return Q(
        Q(consumer_type__app_label=app) &
        Q(consumer_type__model=model.lower()) &
        Q(action_type__in=action_types))



from threading import local

_thread_locals = local()


def get_current_session():
    return getattr(_thread_locals, 'session', None)

def get_current_user():
    return getattr(_thread_locals, 'user', None)

def get_current_ip():
    return getattr(_thread_locals, 'ip', None)


def get_current_request():
    return getattr(_thread_locals, 'request', None)


class ThreadLocals:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        _thread_locals.request = request
        _thread_locals.user = getattr(request, 'user', None)
        _thread_locals.session = getattr(request, 'session', None)
        _thread_locals.ip = request.META.get('HTTP_X_FORWARDED_FOR')
        response = self.get_response(request)

        return response
